{
File: MemoSearch.pas

Description: Allow for easy Search/Replace of a TMemo and its decendants.
             Customizable Search/Replace/Next keys.

History:
    Version 1.2:
        Fixed problem with TRichEdit (Thanks to Jean-Fabien Connault!!!)
        Added Runtime interface via methods Search, Replace, and Next to
        bypass the dialog.
        Added PromptReplace dialog to prompt each replacement.
    Version 1.1:
        Added Replace functionality.
        Fixed minor problem with shortcuts not getting processed correctly.
    Version 1.0:
        Created.
}
unit MemoSearch;

{
How to install the TMemoSearch component in Delphi 5:

1.  Run Delphi and choose Component->Install Component on the menu.
2.  In the "Unit file name", enter the fully qualified path and file
    name of where the MemoSearch.pas unit is, or chose the Browse
    button to find it.
3.  Click OK to recompile the package with the new component.
4.  The TMemoSearch component should now be in the Additional tab of the
    component tabs.

How to use the TMemoSearch component in Delphi 5:

1.  Choose the Additional tab of the components and click on the
    TMemoSearch component.
2.  Drop the component onto the form that contains a TMemo you want to
    attach it to.
3.  Set the "Memo" property to the TMemo component.
4.  If needed, change the SearchKey, ReplaceKey, and NextKey properties
    to the key sequence of choice.
5.  Run the application and set focus to the TMemo.
6.  The Search, Replace, and Next key strokes should be enabled at this point
    to do your searching and replacing. For example, once the TMemo has focus,
    type in (Ctl-F) to bring up the Search dialog.
}

{
Properties:

property SearchKey : TShortCut;
    The key sequence to start a search.  A dropdown is provided to show the
    allowable keys, however, any key sequence can be applied at runtime.

property ReplaceKey : TShortCut;
    The key sequence to start a replace.  A dropdown is provided to show the
    allowable keys, however, any key sequence can be applied at runtime.

property SearchAgainKey : TShortCut;
    The key sequence to continue a search or replace function.  A dropdown is
    provided to show the allowable keys, however, any key sequence can be
    applied at runtime.

property HideAfterSearch : Boolean;
    Property to determine if the Find or Replace dialog is hidden after the Search
    or Replace button is pressed. If False, dialog stays visible.

property Memo : TCustomMemo;
    The Memo we are dealing with here.

property Options : TFindOptions;
    Standard Find/Replace options.  See Delphi help for more info.

property SearchString : string;
    Read-Only. String we are searching for.

property ShowReplaceCount : Boolean;
    Property to determine if we show replacement count when finished.  If True,
    A dialog pops up when all replacements are done showing count. Only shows
    if ReplaceAll was first selected.

property HelpContext: THelpContext;
    If frShowHelp is included in the Options, a Help button is visible on the
    Find/Replace dialog.  When pressed, this property contains the context for
    the help.

property PromptReplace: Boolean
    Prompt before each replace if True.

Methods:

procedure Search(vsSearch : string);
    Calls the search methods by bypassing the Search dialog.

procedure Replace(vsSearch : string);
    Calls the replace methods by bypassing the Replace dialog.

procedure Next;
    Performs Next function of either Search or Replace.

}

interface

uses
  Windows, Messages, Classes, Controls, StdCtrls, Dialogs, Forms, SysUtils,
  Menus;

{(*}
const
  WM_DO_SEARCH = WM_USER + 300;
  WM_DO_REPLACE = WM_USER + 301;
type
  TFindMode = ( fmFind, fmReplace );
  TMemoSearch = class(TComponent)
  private
    FSearchString : string;
    FCurrentPos : Integer;
    FOldPos : Integer;
    FMemo : TCustomMemo;
    FFindDialog : TFindDialog;
    FLocalWndProc : Pointer;
    PrevWndProc : TFarProc;
    FOptions : TFindOptions;
    OHandle : THandle;
    FCloseAfter: Boolean;
    FFindSeq: TShortCut;
    FNextSeq: TShortCut;
    FReplaceSeq: TShortCut;
    FIncludeReplace: Boolean;
    FHelpContext : THelpContext;
    FMode : TFindMode;
    FNumReplace : Integer;
    FShowReplaceCount: Boolean;
    FPromptReplace: Boolean;
    FReplaceAll : Boolean;
    procedure SetCustomMemo(const Value: TCustomMemo);
    procedure LocalWndProc(var Msg: TMessage);
    function FPos( Search, Str : string ): Integer;
    function RPos( Search, Str : string ): Integer;
    procedure SetFindSeq(const Value: TShortCut);
    procedure SetNextSeq(const Value: TShortCut);
    procedure SetHelpContext(const Value: THelpContext);
    function GetHelpContext: THelpContext;
    procedure CreateDialog;
    procedure SetReplaceSeq(const Value: TShortCut);
    procedure Execute(IsReplace: Boolean; SearchString : string = '');
    procedure PreProcessSearch(var Str, Search: string);
    function ShowPromptDialog( X, Y : Integer ) : Integer;
   protected
    procedure FindDialogOnFind(Sender: TObject);
    procedure FindDialogOnReplace(Sender: TObject);
    procedure FindDialogOnClose(Sender: TObject);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure Search(vsSearch : string);
    procedure Replace(vsSearch : string);
    procedure Next;
  published
    property SearchKey : TShortCut read FFindSeq write SetFindSeq;
    property ReplaceKey : TShortCut read FReplaceSeq write SetReplaceSeq;
    property SearchAgainKey : TShortCut read FNextSeq write SetNextSeq;
    property HideAfterSearch : Boolean read FCloseAfter write FCloseAfter default True;
    property Memo : TCustomMemo read FMemo write SetCustomMemo;
    property Options : TFindOptions read FOptions write FOptions;
    property PromptReplace : Boolean read FPromptReplace write FPromptReplace;
    property SearchString : string read FSearchString;
    property ShowReplaceCount : Boolean read FShowReplaceCount write FShowReplaceCount;
    property HelpContext: THelpContext read GetHelpContext write SetHelpContext;
  end;
{*)}

procedure Register;

implementation

const
    FChars : set of Char = [ 'a'..'z', 'A'..'Z', '0'..'9', '_' ];

resourcestring
    sNO_MORE_RESULTS        = 'No more instances of "%s" found.';
    sNO_RESULT_FOUND        = 'Search string "%s" not found.';
    sNUM_REPLACEMENTS       = 'Replaced %d instances of "%s".';
    sSEARCH_CAPTION         = 'Search Results';
    sREPLACE_CAPTION        = 'Replace Results';
    sREPLACE_TEXT           = 'Replace current text?';

procedure Register;
begin
    RegisterComponents ( 'Additional', [ TMemoSearch ] );
end;

{ TMemoSearch }

//-----------------------------------------------------------------------------

constructor TMemoSearch.Create(AOwner: TComponent);
begin
    inherited;

    FFindDialog := nil;
    FSearchString := '';
    FCurrentPos := -1;
    FOldPos := -1;
    OHandle := 0;
    FHelpContext := 0;
    FIncludeReplace := False;
    FOptions := [ frDown ];
    FCloseAfter := True;
    SearchKey := ShortCut( Ord( 'F' ), [ ssCtrl ] );
    SearchAgainKey := VK_F3;
    ReplaceKey := ShortCut( Ord( 'R' ), [ ssCtrl ] );
    FShowReplaceCount := False;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.CreateDialog;
begin
    if Assigned( FFindDialog ) then
        FFindDialog.Free;

    if FIncludeReplace then begin
        FFindDialog := TReplaceDialog.Create( Self );
        TReplaceDialog( FFindDialog ).OnReplace := FindDialogOnReplace;
    end
    else
        FFindDialog := TFindDialog.Create( Self );
    FFindDialog.OnFind := FindDialogOnFind;
    FFindDialog.OnClose := FindDialogOnClose;
    FFindDialog.HelpContext := FHelpContext;
end;

//-----------------------------------------------------------------------------

destructor TMemoSearch.Destroy;
begin
    FFindDialog.Free;

    inherited;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.Execute(IsReplace: Boolean; SearchString : string = '');
begin
    FSearchString := '';
    if Assigned( FFindDialog ) then begin
        FFindDialog.Free;
        FFindDialog := nil;
    end;
    FIncludeReplace := IsReplace;
    FReplaceAll := False;

    CreateDialog;

    FFindDialog.Options := FOptions;

    FNumReplace := 0;
    if SearchString <> '' then
        FFindDialog.FindText := SearchString
    else if FMemo.SelLength > 0 then
        FFindDialog.FindText := FMemo.SelText;

    if SearchString = '' then
        FFindDialog.Execute
    else begin
        if IsReplace then
            FindDialogOnReplace( Self )
        else
            FindDialogOnFind( Self );
    end;

end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.FindDialogOnClose(Sender: TObject);
begin
    FOptions := FFindDialog.Options;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.FindDialogOnFind(Sender: TObject);
var
    Str             : string;
    Search          : string;
begin
    FMode := fmFind;

    PreProcessSearch( Str, Search );

    if FCurrentPos >= 0 then begin
        if FOldPos > -1 then begin
            if ( frDown in FFindDialog.Options ) then
                FCurrentPos := FCurrentPos + FOldPos;
        end;
        FMemo.SelStart := FCurrentPos;
        FMemo.SelLength := Length( FSearchString );
        // 9-17-2000, Jean-Fabien Connault
        SendMessage( FMemo.Handle, EM_SCROLLCARET, 0, 0 );
        FMemo.SetFocus;
        if FCloseAfter then
            PostMessage( FFindDialog.Handle, WM_CLOSE, 0, 0 );
    end;

    if FCurrentPos < 0 then begin
        if FOldPos > -1 then
            MessageBox( Application.Handle, PChar( Format( sNO_MORE_RESULTS, [ FSearchString ] ) ),
                PChar( sSEARCH_CAPTION ), MB_OK )
        else
            MessageBox( Application.Handle, PChar( Format( sNO_RESULT_FOUND, [ FSearchString ] ) ),
                PChar( sSEARCH_CAPTION ), MB_OK );
    end;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.FindDialogOnReplace(Sender: TObject);
var
    Str             : string;
    Search          : string;
    WhatInstance    : integer;
    tmpPos          : integer;
    bDoReplace      : Boolean;
    xPos            : Integer;
    yPos            : Integer;
    xyPoint         : TPoint;
    tmpHide         : Boolean;
    lRes            : LongInt;
begin
    FMode := fmReplace;

    WhatInstance := FNumReplace;

    PreProcessSearch( Str, Search );

    if FCurrentPos >= 0 then begin
        if FOldPos > -1 then begin
            if ( frDown in FFindDialog.Options ) then
                FCurrentPos := FCurrentPos + FOldPos;
        end;
        FMemo.SelStart := FCurrentPos;
        tmpPos := FCurrentPos;
        FMemo.SelLength := Length( FSearchString );
        bDoReplace := True;
        SendMessage( FMemo.Handle, EM_SCROLLCARET, 0, 0 );

        if FPromptReplace then begin
            lRes := SendMessage( FMemo.Handle, Messages.EM_POSFROMCHAR, FMemo.SelStart, 0 );
            xyPoint.X := LoWord( lRes );
            xyPoint.Y := HiWord( lRes );
            xyPoint := FMemo.ClientToScreen( xyPoint );
            xPos := xyPoint.x;
            yPos := xyPoint.y;
            tmpHide := False;
            // unhide if hidden selection
            if TMemo( FMemo ).HideSelection then begin
                TMemo( FMemo ).HideSelection := False;
                tmpHide := True;
            end;
            case ShowPromptDialog( xPos, yPos ) of
                mrNo, mrCancel: begin
                    bDoReplace := False;
                end;
                mrYesToAll: begin
                    FPromptReplace := False;
                    FReplaceAll := True;
                    bDoReplace := True;
                end;
            end;
            if tmpHide then
                TMemo( FMemo ).HideSelection := True;
        end;
        
        if bDoReplace then begin
            SendMessage( FMemo.Handle, EM_REPLACESEL, 1, Longint( PChar( TReplaceDialog( FFindDialog ).ReplaceText ) ) );
            FCurrentPos := FMemo.SelStart;
            Inc( FNumReplace );
            FMemo.SelStart := tmpPos;
            FMemo.SelLength := Length( TReplaceDialog( FFindDialog ).ReplaceText );
        end
        else
            FCurrentPos := FCurrentPos + FMemo.SelLength;
            
        // 9-17-2000, Jean-Fabien Connault
        SendMessage( FMemo.Handle, EM_SCROLLCARET, 0, 0 );
        
        if FCloseAfter and ( FNumReplace = 1 ) then
            PostMessage( FFindDialog.Handle, WM_CLOSE, 0, 0 );
        if FReplaceAll then
            FindDialogOnReplace( Sender );
    end;

    if ( FCurrentPos < 0 ) then begin
        if ( FNumReplace = 0 ) or ( not FReplaceAll ) then begin
            if FOldPos > -1 then
                MessageBox( Application.Handle, PChar( Format( sNO_MORE_RESULTS, [ FSearchString ] ) ),
                    PChar( sREPLACE_CAPTION ), MB_OK )
            else
                MessageBox( Application.Handle, PChar( Format( sNO_RESULT_FOUND, [ FSearchString ] ) ),
                    PChar( sREPLACE_CAPTION ), MB_OK );
        end
        else if ( WhatInstance = 0 ) and FShowReplaceCount and ( FReplaceAll ) then
            MessageBox( Application.Handle, PChar( Format( sNUM_REPLACEMENTS, [ FNumReplace, FSearchString ] ) ),
                PChar( sREPLACE_CAPTION ), MB_OK );
    end;
end;

//-----------------------------------------------------------------------------

function TMemoSearch.FPos(Search, Str: string): Integer;
var
    Start, Ending   : Boolean;
    NotFound        : Boolean;
    OldPos          : Integer;
begin
    NotFound := False;
    OldPos := 0;
    repeat
        Result := Pos( Search, Str );

        if ( Result > 0 ) and ( frWholeWord in FFindDialog.Options ) then begin
            Start := ( Result = 1 ) or ( not ( Str[ Result - 1 ] in FChars ) );
            Ending := ( Result + Length( Search ) - 1 = Length( Str ) ) or
                ( not ( Str[ Result + Length( Search ) ] in FChars ) );

            if not ( Start and Ending ) then begin
                OldPos := OldPos + ( Result + Length( Search ) ) - 1;
                Str := Copy( Str, Result + Length( Search ), Length( Str ) );
                Result := 0;
            end;
            if Str = '' then begin
                OldPos := 0;
                NotFound := True;
            end;
        end
        else begin
            OldPos := 0;
            NotFound := True;
        end;
    until ( Result > 0 ) or ( NotFound );

    if OldPos > 0 then
        Result := Result + OldPos;
end;
//-----------------------------------------------------------------------------

function TMemoSearch.GetHelpContext: THelpContext;
begin
    Result := FHelpContext;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.LocalWndProc(var Msg: TMessage);
var
    ShortCut        : TShortCut;
const
    InProcess       : Boolean = False;
begin
    case Msg.Msg of
        WM_DO_SEARCH: begin
            Application.ProcessMessages;
            Search( '' );
            Msg.Result := 0;
            Exit;
        end;
        WM_DO_REPLACE: begin
            Application.ProcessMessages;
            Replace( '' );
            Msg.Result := 0;
            Exit;
        end;
        WM_KEYDOWN: begin
            ShortCut := Byte( TWMKey( Msg ).CharCode );
            if GetKeyState( VK_SHIFT ) < 0 then Inc( ShortCut, scShift );
            if GetKeyState( VK_CONTROL ) < 0 then Inc( ShortCut, scCtrl );
            if GetKeyState( VK_MENU ) < 0 then Inc( ShortCut, scAlt );
            if ShortCut = FFindSeq then begin
                PostMessage( FMemo.Handle, WM_DO_SEARCH, 0, 0 );
                InProcess := True;
                Exit;
            end
            else if ShortCut = FReplaceSeq then begin
                PostMessage( FMemo.Handle, WM_DO_REPLACE, 0, 0 );
                InProcess := True;
                Exit;
            end
            else if ( ShortCut = FNextSeq ) then begin
                if ( FSearchString <> '' ) then begin
                    if FMode = fmFind then
                        FindDialogOnFind( Self )
                    else
                        FindDialogOnReplace( Self );
                    InProcess := True;
                    Exit;
                end;
            end;
        end;
        WM_CHAR: begin
            if InProcess then begin
                InProcess := False;
                TWMKey( Msg ).CharCode := 0;
                Msg.Result := 1;
                Exit;
            end;
        end;
    end;

    Msg.Result := CallWindowProc( PrevWndProc, OHandle, Msg.Msg, Msg.wParam, Msg.lParam );
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.Next;
begin
    if FMode = fmFind then
        FindDialogOnFind( Self )
    else
        FindDialogOnReplace( Self );
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.Notification(AComponent: TComponent; Operation: TOperation);
begin
    inherited;

    if ( FMemo <> nil ) and ( AComponent = FMemo ) and ( Operation = opRemove ) then begin
        if OHandle <> 0 then begin
            SetWindowLong( OHandle, GWL_WNDPROC, LongInt( PrevWndProc ) );
            OHandle := 0;
        end;
        FMemo := nil;
    end;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.PreProcessSearch(var Str, Search: string);
begin
    if FSearchString <> FFindDialog.FindText then begin
        FCurrentPos := -1;
        FOldPos := -1;
        FSearchString := FFindDialog.FindText;
        FReplaceAll := frReplaceAll in FFindDialog.Options;
    end
    else begin
        FOldPos := FMemo.SelStart;
        if ( frDown in FFindDialog.Options ) then
            FOldPos := FOldPos + FMemo.SelLength;
    end;

    Str := FMemo.Lines.Text;
    Search := FSearchString;

    if ( FOldPos > -1 ) then
        if ( frDown in FFindDialog.Options ) then
            Str := Copy( Str, FOldPos + 1, Length( Str ) )
        else
            Str := Copy( Str, 1, FOldPos );

    if not ( frMatchCase in FFindDialog.Options ) then begin
        Str := UpperCase( Str );
        Search := UpperCase( Search );
    end;

    if ( frDown in FFindDialog.Options ) then
        FCurrentPos := FPos( Search, Str ) - 1
    else
        FCurrentPos := RPos( Search, Str ) - 1;

    if FCloseAfter then
        PostMessage( FFindDialog.Handle, WM_CLOSE, 0, 0 );
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.Replace(vsSearch : string);
begin
    Execute( True, vsSearch );
end;


//-----------------------------------------------------------------------------

function TMemoSearch.RPos(Search, Str: string): Integer;
var
    Start, Ending   : Boolean;
    NotFound        : Boolean;
begin
    NotFound := False;
    repeat
        Result := Length( Str ) - Length( Search ) + 1;
        while ( Result > 0 ) do begin
            if Copy( Str, Result, Length( Search ) ) = Search then
                Break;
            Dec( Result );
        end;

        if ( Result > 0 ) and ( frWholeWord in FFindDialog.Options ) then begin
            Start := ( Result = 1 ) or ( not ( Str[ Result - 1 ] in FChars ) );
            Ending := ( Result + Length( Search ) - 1 = Length( Str ) ) or
                ( not ( Str[ Result + Length( Search ) ] in FChars ) );

            if not ( Start and Ending ) then begin
                Str := Copy( Str, 1, Result );
                Result := 0;
            end;
            if Str = '' then
                NotFound := True;
        end
        else begin
            NotFound := True;
        end;
    until ( Result > 0 ) or ( NotFound );
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.Search( vsSearch : string );
begin
    Execute( False, vsSearch );
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.SetCustomMemo(const Value: TCustomMemo);
begin
    if ( FMemo <> Value ) then begin
        if OHandle <> 0 then begin
            SetWindowLong( OHandle, GWL_WNDPROC, LongInt( PrevWndProc ) );
            OHandle := 0;
        end;

        FMemo := Value;
        FMemo.FreeNotification( Self );
    end;

    if FMemo <> nil then begin
        FLocalWndProc := MakeObjectInstance( LocalWndProc );
        OHandle := Value.Handle;
        PrevWndProc := TFarProc( GetWindowLong( OHandle, GWL_WNDPROC ) );
        SetWindowLong( OHandle, GWL_WNDPROC, LongInt( FLocalWndProc ) );
    end;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.SetFindSeq(const Value: TShortCut);
begin
    FFindSeq := Value;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.SetHelpContext(const Value: THelpContext);
begin
    FHelpContext := Value;
    if Assigned( FFindDialog ) then
        FFindDialog.HelpContext := Value;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.SetNextSeq(const Value: TShortCut);
begin
    FNextSeq := Value;
end;

//-----------------------------------------------------------------------------

procedure TMemoSearch.SetReplaceSeq(const Value: TShortCut);
begin
  FReplaceSeq := Value;
end;

//-----------------------------------------------------------------------------

function TMemoSearch.ShowPromptDialog( X, Y : Integer): Integer;
begin
  with CreateMessageDialog( sREPLACE_TEXT, mtConfirmation, [ mbYes, mbNo, mbYesToAll ] ) do
    try
      HelpContext := 0;
      if X >= 0 then Left := X;
      if Left + Width > Screen.Width then
        Left := Screen.Width - Width;
      if Y >= 0 then Top := Y + GetParentForm( FMemo ).Canvas.TextHeight( '0' ) + 5;
      if Top + Height > Screen.Height then
        Top := Y - Height - 5;
      Result := ShowModal;
    finally
      Free;
    end;

end;

end.
